/* Bluetooth stack headers */
#include "bg_types.h"
#include "native_gecko.h"
#include "gatt_db.h"
#include "em_i2c.h"
#include "em_cmu.h"
#include "app.h"

#define TIMER_ID	1

/* Print boot message */
static void bootMessage(struct gecko_msg_system_boot_evt_t *bootevt);
static void setAdvScanData(void);
static void I2C1_init(void);
static void SI7021_getData(void);

/* Flag for indicating DFU Reset must be performed */
static uint8_t boot_to_dfu = 0;
static customAdv_t advData;
static customScanResp_t scanData;
static uint8_t conn_handle;

/* Main application */
void appMain(gecko_configuration_t *pconfig)
{
#if DISABLE_SLEEP > 0
  pconfig->sleep.flags = 0;
#endif

  /* Initialize debug prints. Note: debug prints are off by default. See DEBUG_LEVEL in app.h */
  initLog();
  setAdvScanData();
  I2C1_init();
  GPIO_PinModeSet(BSP_LED0_PORT, BSP_LED0_PIN, gpioModePushPull, 0);

  /* Initialize stack */
  gecko_init(pconfig);

  while (1) {
    /* Event pointer for handling events */
    struct gecko_cmd_packet* evt;

    /* if there are no events pending then the next call to gecko_wait_event() may cause
     * device go to deep sleep. Make sure that debug prints are flushed before going to sleep */
    if (!gecko_event_pending()) {
      flushLog();
    }

    /* Check for stack event. This is a blocking event listener. If you want non-blocking please see UG136. */
    evt = gecko_wait_event();

    /* Handle events */
    switch (BGLIB_MSG_ID(evt->header)) {
      /* This boot event is generated when the system boots up after reset.
       * Do not call any stack commands before receiving the boot event.
       * Here the system is set to start advertising immediately after boot procedure. */
      case gecko_evt_system_boot_id:

        bootMessage(&(evt->data.evt_system_boot));
        printLog("boot event - starting advertising\r\n");

        /* Set advertising parameters. 100ms advertisement interval.
         * The first parameter is advertising set handle
         * The next two parameters are minimum and maximum advertising interval, both in
         * units of (milliseconds * 1.6).
         * The last two parameters are duration and maxevents left as default. */
        gecko_cmd_le_gap_set_advertise_timing(0, 160, 160, 0, 0);
        gecko_cmd_le_gap_bt5_set_adv_data(0, 0, sizeof(advData), (const uint8_t*)&advData);
        gecko_cmd_le_gap_bt5_set_adv_data(0, 1, sizeof(scanData), (const uint8_t*)&scanData);
        /* Start general advertising and enable connections. */
        gecko_cmd_le_gap_start_advertising(0, le_gap_user_data, le_gap_connectable_scannable);
        break;

      case gecko_evt_le_connection_opened_id:
    	conn_handle = evt->data.evt_le_connection_opened.connection;
    	GPIO_PinOutSet(BSP_LED0_PORT, BSP_LED0_PIN);
        printLog("connection opened\r\n");
        gecko_cmd_hardware_set_soft_timer(32768, TIMER_ID, 0);
      break;

      case gecko_evt_le_connection_closed_id:
    	GPIO_PinOutClear(BSP_LED0_PORT, BSP_LED0_PIN);
    	gecko_cmd_hardware_set_soft_timer(0, TIMER_ID, 0);		// stop measurement timer
        printLog("connection closed, reason: 0x%2.2x\r\n", evt->data.evt_le_connection_closed.reason);
        /* Check if need to boot to OTA DFU mode */
        if (boot_to_dfu) {
          /* Enter to OTA DFU mode */
          gecko_cmd_system_reset(2);
        } else {
          /* Restart advertising after client has disconnected */
          gecko_cmd_le_gap_start_advertising(0, le_gap_general_discoverable, le_gap_connectable_scannable);
        }
      break;

      /* Events related to OTA upgrading
         ----------------------------------------------------------------------------- */

      /* Check if the user-type OTA Control Characteristic was written.
       * If ota_control was written, boot the device into Device Firmware Upgrade (DFU) mode. */
      case gecko_evt_gatt_server_user_write_request_id:
        if (evt->data.evt_gatt_server_user_write_request.characteristic == gattdb_ota_control) {
          /* Set flag to enter to OTA mode */
          boot_to_dfu = 1;
          /* Send response to Write Request */
          gecko_cmd_gatt_server_send_user_write_response(
            evt->data.evt_gatt_server_user_write_request.connection,
            gattdb_ota_control,
            bg_err_success);

          /* Close connection to enter to DFU OTA mode */
          gecko_cmd_le_connection_close(evt->data.evt_gatt_server_user_write_request.connection);
        }
      break;

      case gecko_evt_hardware_soft_timer_id:
    	  if (evt->data.evt_hardware_soft_timer.handle == TIMER_ID) {
    		  SI7021_getData();
    	  }
      break;

      default:
        break;
    }
  }
}

/* Print stack version and local Bluetooth address as boot message */
static void bootMessage(struct gecko_msg_system_boot_evt_t *bootevt)
{
#if DEBUG_LEVEL
  bd_addr local_addr;
  int i;

  printLog("stack version: %u.%u.%u\r\n", bootevt->major, bootevt->minor, bootevt->patch);
  local_addr = gecko_cmd_system_get_bt_address()->address;

  printLog("local BT device address: ");
  for (i = 0; i < 5; i++) {
    printLog("%2.2x:", local_addr.addr[5 - i]);
  }
  printLog("%2.2x\r\n", local_addr.addr[0]);
#endif
}

void setAdvScanData(void)
{
	advData.len_flags = 0x02;
	advData.type_flags = 0x01;
	advData.val_flags = 0x06;					// general discoverable mode

	advData.len_manuf = 15;
	advData.type_manuf = 0xFF;
	advData.company_LO = 0xFF;					// Silabs company ID (0x02FF)
	advData.company_HI = 0x02;
	strncpy(advData.manuf_name, "Silicon Labs", 12);

	scanData.len_name = 8;
	scanData.type_name = 9;						// full device name
	strncpy(scanData.name, "CSCI470", 7);
}

static void I2C1_init(void)
{
	I2C_Init_TypeDef i2cInit = I2C_INIT_DEFAULT;
	CMU_ClockEnable(cmuClock_HFPER, true);		// enable clocks
	CMU_ClockEnable(cmuClock_I2C1, true);
	GPIO_PinModeSet(BSP_I2C1_SCL_PORT, BSP_I2C1_SCL_PIN, gpioModeWiredAnd, 1);
	GPIO_PinModeSet(BSP_I2C1_SDA_PORT, BSP_I2C1_SDA_PIN, gpioModeWiredAnd, 1);
	I2C1->ROUTELOC0 = BSP_I2C1_SDA_LOC + (BSP_I2C1_SCL_LOC << 8);
	I2C1->ROUTEPEN = 3;
	I2C_Init(I2C1, &i2cInit);
	GPIO_PinModeSet(ENV_SENSE_ENABLE_PORT, ENV_SENSE_ENABLE_PIN, gpioModePushPull, 1);
}

static void SI7021_getData(void)
{
	// configure I2C transactions
	I2C_TransferReturn_TypeDef ret;
	I2C_TransferSeq_TypeDef seq;
	uint8_t humi;
	int8_t  temp;
	uint8_t dataW[1] = {0xF5};					// command to measure temp & humi
	uint8_t dataR[2];
	seq.addr = 0x80;							// sensor slave address
	seq.flags = I2C_FLAG_WRITE;
	seq.buf[0].data = dataW;
	seq.buf[0].len = 1;

	// request data measurement
	ret = I2C_TransferInit(I2C1, &seq);
	while (ret == i2cTransferInProgress)
		ret = I2C_Transfer(I2C1);

	// get sensor data
	seq.flags = I2C_FLAG_READ;
	seq.buf[0].data = dataR;
	seq.buf[0].len = 2;							// read 2 bytes of humi
	do {
		ret = I2C_TransferInit(I2C1, &seq);
		while (ret == i2cTransferInProgress)
			ret = I2C_Transfer(I2C1);
	}
	while (ret == i2cTransferNack);				// wait for data ready

	while (ret == i2cTransferInProgress)
		ret = I2C_Transfer(I2C1);
	humi = (((((dataR[0] << 8) + dataR[1])*125 + 0x8000) >> 16) - 6);	// compute humi
	gecko_cmd_gatt_server_write_attribute_value(gattdb_humi_value, 0, 1, (uint8_t*)&humi);

	if (humi > 99)								// saturate humi value
		humi = 99;

	dataW[0] = 0xE0;							// get temp from the last humi conversion
	seq.flags = I2C_FLAG_WRITE_READ;
	seq.buf[0].data = dataW;
	seq.buf[0].len = 1;							// write 1 byte of command
	seq.buf[1].data = dataR;
	seq.buf[1].len = 2;							// read 2 bytes of temp
	ret = I2C_TransferInit(I2C1, &seq);
	while (ret == i2cTransferInProgress)
		ret = I2C_Transfer(I2C1);
	temp = ((((dataR[0] << 8) + dataR[1])*176 + 0x8000) >> 16) - 47;	// compute temp in Celcius
	gecko_cmd_gatt_server_write_attribute_value(gattdb_temp_value, 0, 1, (uint8_t*)&temp);
}
